FILE_NAME = 'etc/yas/yas.yml'

PARAMETERS = dict(
    ignored_types=['desktop_notification', 'user_typing'],
    slack_app_token=None,
    bot_name=None,
    handler_list=['yas.default_handler'],
    debug=False,
    default_response='Sure...write some more code then I can do that!',
    log_level='WARN'
)



import hashlib

from slackclient import SlackClient

from yas import SlackClientFailure, NoBot, HandlerError
from yas.handler_manager import HandlerManager
from yas.logging import logger
from yas.yaml_file_config import YamlConfiguration


config = YamlConfiguration()

def hash(string):
    return hashlib.md5(str(string).encode('utf-8')).hexdigest()

class Client(SlackClient):

    def __init__(self, data_filter=None, ignored_types=None):
        super().__init__(config.slack_app_token)

        self.bot_id = self.__retrieve_bot_user_id()

        self.at_bot = "<@" + self.bot_id + ">"
        self.data_filter = data_filter

        self.ignored_types = ignored_types or config.ignored_types

        self.handler_manager = HandlerManager(config.handler_list, debug=config.debug)

    def __retrieve_bot_user_id(self):
        logger.log.info("Retrieving users list for self identification...")
        api_call = self.api_call("users.list")
        if api_call.get('ok'):
            # retrieve all users so we can find our bot
            users = api_call.get('members')
            for user in users:
                if 'name' in user and user.get('name') == config.bot_name:
                    return user.get('id')
            else:
                raise NoBot(config.bot_name)
        else:
            raise SlackClientFailure("Unable to query api")

    def process_changes(self, data):
        super().process_changes(data)
        logger.log.debug(f"Raw data: {data}")
        if self.data_filter(self, data):
            channel = data.get('channel')
            def reply(message, channel=channel, thread=None, reply_broadcast=None):
                self.rtm_send_message(channel, message, thread, reply_broadcast)
            data['yas_hash'] = hash(data)
            logger.log.info(f"Processing: {data}")
            try:
                self.handler_manager.handle(data, reply)
            except Exception as exception:
                reply(f"Err, sorry, that threw an exception: {exception}. Try again or reach out to the maintainers.")

    def listen(self):
        if self.rtm_connect():
            logger.log.info("Slack bot connected as {}<{}> and running!".format(config.bot_name, self.bot_id))
            while True:
                self.rtm_read()
        else:
            logger.log.fatal("Connection failed. Invalid Slack token or bot ID?")



import os
import sys

import yaml

from . import configuration


class YamlConfigError(Exception):
    pass

class RequiredParameter(YamlConfigError):
    '''{parameter} is a required configuration parameter and must be set to a non empty value. Please correct this in your configuration file at {config_file}.'''
    def __init__(self, parameter, config_file):
        self.message = self.__doc__.format(parameter=parameter, config_file=config_file)

class YamlConfiguration(object):

    def __parse_config(self, config_file_path):
        with open(config_file_path, 'r') as config_file:
            raw_config = yaml.load(config_file)

            parameters = configuration.PARAMETERS
            for parameter in parameters:
                default = parameters[parameter]
                value = raw_config.get(parameter, default)

                if not value and default is None:
                    raise RequiredParameter(parameter, config_file_path)

                setattr(self, parameter, value)

    def __init__(self):
        config_file = os.path.join(sys.prefix, configuration.FILE_NAME)
        self.__parse_config(config_file)



from yas.yaml_file_config import YamlConfiguration
from yas.logging import logger, log


class YasHandler:
    def __init__(self, log=log):
        self.log = log

class YasError(Exception):
    def __init__(self, msg):
        logger.log.fatal(msg)

class HandlerError(YasError):
    def __init__(msg):
        super().__init__(msg)

class SlackClientFailure(YasError):
    def __init__(self, msg):
        super().__init__(msg)

class NoBot(SlackClientFailure):
    def __init__(self, bot_name):
        super().__init__(f"Could not find bot user with the name {bot_name}, please check your yas config.")

class NotAHandler(HandlerError):
    def __init__(self, not_a_handler):
        super().__init__(f"{not_a_handler} is not a handler.")



import imp
import inspect
import sys

from yas import YasHandler, NotAHandler, HandlerError
from yas.logging import logger, log


def import_from_dotted_path(dotted_names, path=None):
    """ import_from_dotted_path('foo.bar') -> from foo import bar; return bar

    This function borrowed from Chase Seibert at
    http://chase-seibert.github.io/blog/2014/04/23/python-imp-examples.html
    """
    next_module, remaining_names = dotted_names.split('.', 1)
    file, pathname, description = imp.find_module(next_module, path)
    module = imp.load_module(next_module, file, pathname, description)

    if hasattr(module, remaining_names):
        return getattr(module, remaining_names)

    if '.' not in remaining_names:
        return module

    return import_from_dotted_path(remaining_names, path=module.__path__)

def get_ancestors(target_class):
    bases = []
    if target_class is not object:
        for base in target_class.__bases__:
            bases.extend(get_ancestors(base))

    bases.extend(target_class.__bases__)
    return bases

def is_handler(test_class):
    logger.log.info(f"Checking if {test_class} is a handler")
    public_methods = [member[0] for member in inspect.getmembers(test_class)
                      if not member[0].startswith('__')]

    test_class_bases = get_ancestors(test_class)
    logger.log.debug(f"{test_class} has bases {test_class_bases}")

    class_derives_yas_handler = str(YasHandler) in [str(base) for base in test_class_bases]
    class_defines_required_methods = 'handle' in public_methods and 'test' in public_methods

    check_or_x = {True: '✔', False: '✘'}

    logger.log.info(f"{check_or_x[class_derives_yas_handler]} {test_class} derives {YasHandler}")
    logger.log.info(f"{check_or_x[class_defines_required_methods]} {test_class} defines required methods")

    return class_derives_yas_handler and class_defines_required_methods

def find_handlers_in_module(handler_module):
    return [handler[1](log=log) for handler in inspect.getmembers(handler_module, inspect.isclass)
            if is_handler(handler[1])]

class HandlerManager:
    def __init__(self, handler_list, debug=False):

        logger.log.info(f"Loading handlers from {handler_list}")

        self.handler_list = []
        for handler_name in handler_list:
            logger.log.info(f"Searching for {handler_name}")
            try:

                handler = import_from_dotted_path(handler_name)
                if type(handler) is type and is_handler(handler):
                    logger.log.info(f"Found handler {handler}.")
                    self.handler_list.append(handler(log=log))
                    continue

                module_handlers = find_handlers_in_module(handler)
                if module_handlers:
                    logger.log.info(f"Found handlers in {handler_name}: {module_handlers}")
                    self.handler_list.extend(module_handlers)
                    continue

                raise NotAHandler(handler_name)

            except Exception as exception:
                logger.log.warn(f'Failed to load handler {handler_name}, caught: {exception}')
                if debug:
                    raise exception

        logger.log.info(f'Loaded handlers: {self.handler_list}')

    def handle(self, data, reply):
        logger.log.info(f"Handling {data['yas_hash']}")
        for handler in self.handler_list:
            logger.log.info(f"Testing {data['yas_hash']} against {handler}")
            if handler.test(data):
                logger.log.info(f"Handling {data['yas_hash']} with {handler}")
                handler.handle(data, reply)
                break
        else:
            logger.log.info(f'No handler found for {data}')



from yas.client import Client


def default_data_filter(yas_client, data):
    if data.get('type') not in yas_client.ignored_types \
            and 'channel' in data \
            and data.get('user') != yas_client.bot_id:
        channel = data['channel']
        channel_info = yas_client.api_call('channels.info', channel=channel)
        if channel_info.get('ok') and yas_client.at_bot in data['text']:
            return True
        group_info = yas_client.api_call('groups.info', channel=channel)
        if not channel_info.get('ok', False) and not group_info.get('ok', False):
            return True

def run():
    client = Client(data_filter=default_data_filter)
    client.listen()



from yas import YasHandler, HandlerError
from yas.yaml_file_config import YamlConfiguration


config = YamlConfiguration()

class DefaultHandler(YasHandler):
    '''
    YasHandlers, when registered with a Yas installation in the active yas.yml,
    are tested against, and upon matching applied to, incoming messages by Yas
    as they are receieved.
    '''

    def test(self, data):
        '''
        If this method returns truthy, this class's handle method is called with the data.
        By default, this handler is registered with yas. To disable, remove
        yas.handler from the handler_list in your yas.yml.

        Arguments:

        data :: dict :: An event from the slack RTM websocket. To see the raw
            stream, enable the raw_response handler. The objects are pre filtered by
            the yas client's data_filter method, which by default filters out any
            ignored types from the yas config, objects not associated with a
            channel, and objects originating from the bot. The implementation of this
            behaviour is likely to change in the future, but the default filters are not.
        '''
        return True

    def handle(self, data, reply):
        '''
        Handles a matched data object. If a HandlerError is raised here, it will be caught,
        and its message sent to the originating channel; other errors are not caught and
        will crash yas, which shouldn't be a problem if one of the included daemon scripts
        is used as they will restart yas after a crash.

        Arguments:

        data :: dict :: See test.

        reply :: function(response, channel=channel, thread=None, reply_broadcast=None) ::
            A shortcut function to reply to a message. The response is your message back
            to the channel. The channel, which defaults to the data's channel, is where
            the message will be delivered; besides the default, another useful channel
            is the data's originating user, which would direct message that user. The
            thread is the parent message ID, if you wish to initiate a thread.
            The reply_broadcast argument takes a boolean value and indicates whether a
            threaded message should also reply in the parent channel.

        api_call :: function(method, timeout=None, **kwargs) :: The api_call method of
            the yas client listening for messages. This is provided so that handler
            authors may shoot themselves in the foot. Available methods may be found
            at https://api.slack.com/methods. This function may be restricted or
            removed in the future as it is rather dangerous.

        handler_manager :: yas.handler.HandlerManager :: The HandlerManager instance
            the YasHandler instance is bound to. This is provided so that query response
            listeners may be registered. This argument may be replaced in the future with
            something more explicit.
        '''
        reply(config.default_response)



import logging
from systemd import journal
from yas.yaml_file_config import YamlConfiguration


config = YamlConfiguration()

class Logger:

    def __init__(self, log_level='WARNING'):
        self.log = logging.getLogger('yas')
        self.log.addHandler(journal.JournaldLogHandler())
        self.log.setLevel(logging._nameToLevel[log_level])

logger = Logger(config.log_level)

def log(level_str, msg):
    level = logging._nameToLevel[level_str]
    logger.log.log(level, msg)
